"""A Python package for working with the BIDS schema.

.. autodata:: __version__
.. autodata:: __bids_version__
"""

__version__: str
__bids_version__: str

__all__ = ("__version__", "__bids_version__")


def __getattr__(attr: str) -> str:
    if attr in __all__:
        from . import _version

        globals()[attr] = getattr(_version, attr)
        return globals()[attr]

    raise AttributeError(f"module {__spec__.name!r} has no attribute {attr!r}")



from __future__ import annotations

import io
import re

import pandas as pd
from markdown_it import MarkdownIt
from tabulate import tabulate

from ..utils import WarningsFilter, in_context
from .utils import propagate_fence_exception


@propagate_fence_exception
@in_context(WarningsFilter(("error",)))
def fence(
    source: str,
    language: str,
    css_class: str,
    options: dict,
    md,
    classes: list[str],
    id_value="",
    attrs: dict | None = None,
    **kwargs,
) -> str:
    attrs = attrs or {}
    linenums = attrs.get("linenums", "0") == "1"

    if language == "tsvgz" and "header" not in attrs:
        attrs["noheader"] = True

    classes[:0] = ["tsv-table", "index" if linenums else "noindex"]

    df = pd.read_csv(
        io.StringIO(source),
        sep="\t",
        dtype=str,
        index_col=False,
        keep_default_na=False,
        header=None if "noheader" in attrs else "infer",
    )
    md_table = tabulate(
        df,  # type: ignore
        tablefmt="github",
        showindex=linenums,
        headers="keys",
        numalign="right",
    )
    html = MarkdownIt("commonmark").enable("table").render(md_table)
    if "noheader" in attrs:
        html = re.sub("<thead>.+</thead>", "", html, flags=re.DOTALL)

    html = html.replace("<table>", f'<table class="{" ".join(classes)}">')

    # Remove newlines from HTML to prevent copy-paste from inserting spaces
    return html.replace("\n", "")



"""Utility functions for the bids-specification schema."""

from __future__ import annotations

import logging
import os
import sys
import warnings
from functools import wraps

from . import _lazytypes as lt
from . import data

if lt.TYPE_CHECKING:

    class ValidatorKwargs(lt.TypedDict):
        """Type for the keyword arguments used to create a JSON schema validator."""

        format_checker: lt.NotRequired[lt.FormatChecker]


def get_bundled_schema_path() -> str:
    """Get the path to the schema directory.

    Returns
    -------
    str
        Absolute path to the directory containing schema-related files.
    """
    return str(data.load("schema"))


def get_logger(name: str | None = None, level: int | str | None = None) -> logging.Logger:
    """Return a logger to use.

    Parameters
    ----------
    name : None or str, optional
        Name of the logger. Defaults to None.
    level: None or int, optional
        Level to set for the logger. Defaults to None.
        BIDS_SCHEMA_LOG_LEVEL environment variable can
        be used to set the level and will overwrite
        this value.

    Returns
    -------
    logging.Logger
        logger object.
    """
    logger = logging.getLogger("bidsschematools" + (f".{name}" if name else ""))
    # If explicitly instructed via env var -- set log level
    if log_level := os.getenv("BIDS_SCHEMA_LOG_LEVEL", level):
        set_logger_level(logger, log_level)
    return logger


def configure_logger(lgr: logging.Logger) -> None:
    """Configuring formatting and stream handler for the logger.

    Should not be used when bidsschematools is used as a library.

    Parameters
    ----------
    lgr: logging.Logger

    """
    if len(lgr.handlers) == 0:
        # add a handler if there isn't one
        ch = logging.StreamHandler(stream=sys.stderr)
        lgr.addHandler(ch)
    # Set the formatter for the handlers
    for lh in lgr.handlers:
        lh.setFormatter(logging.Formatter("%(asctime)-15s [%(levelname)8s] %(message)s"))


def set_logger_level(lgr: logging.Logger, level: int | str) -> None:
    """Set the logger level.

    Parameters
    ----------
    lgr : logging.Logger
        logger object for which to change the level.
    level : int or str
        The desired level for the logger.
    """
    if isinstance(level, int):
        pass
    elif level.isnumeric():
        level = int(level)
    elif level.isalpha():
        level = getattr(logging, level)
    else:
        lgr.warning("Do not know how to treat loglevel %s", level)
        return
    lgr.setLevel(level)


def jsonschema_validator(
    schema: dict[str, lt.Any],
    *,
    check_format: bool,
    default_cls: type[lt.JsonschemaValidator] | None = None,
) -> lt.JsonschemaValidator:
    """
    Create a jsonschema validator appropriate for validating instances against a given
    JSON schema

    Parameters
    ----------
    schema : dict[str, Any]
        The JSON schema to validate against
    check_format : bool
        Indicates whether to check the format against format specifications in the
        schema
    default_cls : type[JsonschemaValidator] or None, optional
        The default JSON schema validator class to use to create the
        validator should the appropriate validator class cannot be determined based on
        the schema (by assessing the `$schema` property). If `None`, the class
        representing the latest JSON schema draft supported by the `jsonschema` package

    Returns
    -------
    JsonschemaValidator
        The JSON schema validator

    Raises
    ------
    jsonschema.exceptions.SchemaError
        If the JSON schema is invalid
    """
    try:
        from jsonschema.validators import validator_for
    except ImportError as e:
        raise RuntimeError(
            "The `jsonschema` package is required to validate schemas. "
            "Please install it with `pip install jsonschema`."
        ) from e

    cls_kwargs = {} if default_cls is None else {"default": default_cls}
    # Retrieve appropriate validator class for validating the given schema
    validator_cls = validator_for(schema, **cls_kwargs)

    # Ensure the schema is valid
    validator_cls.check_schema(schema)

    validator_kwargs: ValidatorKwargs
    validator_kwargs = {"format_checker": validator_cls.FORMAT_CHECKER} if check_format else {}
    return validator_cls(schema, **validator_kwargs)  # type: ignore[call-arg]


def in_context(
    context_manager: lt.AbstractContextManager,
) -> lt.Callable[[lt.Callable], lt.Callable]:
    """Convert a context manager into a function decorator.

    Parameters
    ----------
    context_manager : context manager
        The context manager to use.

    Returns
    -------
    Callable
        The function decorator.
    """

    def decorator(func: lt.Callable) -> lt.Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            with context_manager:
                return func(*args, **kwargs)

        return wrapper

    return decorator


class WarningsFilter:
    """Context manager to apply warning filters.

    Arguments are lists of positional arguments to :func:`warnings.filterwarnings`.
    """

    # Only using one positional arg for now. This type can get more complex.
    def __init__(
        self,
        *filters: tuple[
            lt.Literal["default", "error", "ignore", "always", "all", "module", "once"]
        ],
    ) -> None:
        self.filters = filters

    def __enter__(self) -> None:
        self.catcher = warnings.catch_warnings()
        self.catcher.__enter__()
        for filt in self.filters:
            warnings.filterwarnings(*filt)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.catcher.__exit__(exc_type, exc_value, traceback)



"""Typing symbols for the bidsschematools package.

This module aims to make the types used by the bidsschematools package available for
type checking without requiring the original packages to be imported at runtime.

The expected usage is::

    from __future__ import annotations

    import bidsschematools._lazytypes as lt

    def func(arg: lt.Any) -> ReturnType:
        ...


If used outside of type annotations, for example, in an ``isinstance()`` check,
the attribute will trigger an import.
"""

import sys

__all__ = (
    # typing
    "Any",
    "Callable",
    "Literal",
    "NotRequired",
    "Protocol",
    "Self",
    "TYPE_CHECKING",
    "TypeVar",
    "TypedDict",
    # collections.abc
    "Iterator",
    "Mapping",
    # contextlib
    "AbstractContextManager",
    # third-party
    "FormatChecker",
    "JsonschemaValidator",
    "Traversable",
)

_type_map = {
    "AbstractContextManager": ("contextlib", "AbstractContextManager"),
    "Iterator": ("collections.abc", "Iterator"),
    "Mapping": ("collections.abc", "Mapping"),
    "Traversable": ("acres.typ", "Traversable"),
    "FormatChecker": ("jsonschema", "FormatChecker"),
    "JsonschemaValidator": ("jsonschema.protocols", "JsonschemaValidator"),
}

RUNTIME_IMPORT: bool

TYPE_CHECKING = False
if TYPE_CHECKING or "sphinx.ext.autodoc" in sys.modules:  # pragma: no cover
    from collections.abc import Callable, Iterator, Mapping
    from contextlib import AbstractContextManager
    from typing import Any, Literal, Protocol, TypedDict, TypeVar

    # Missing imports in 3.10. Avoid breaking third-party autodoc generation.
    try:
        from typing import NotRequired, Self
    except ImportError:  # PY310
        NotRequired = tuple  # type: ignore[misc,assignment]
        Self = object  # type: ignore[misc,assignment]

    from acres.typ import Traversable

    # Optional dependency, avoid breaking third-party autodoc generation.
    try:
        from jsonschema import FormatChecker
        from jsonschema.protocols import Validator as JsonschemaValidator
    except ImportError:
        FormatChecker = object  # type: ignore[misc,assignment]
        JsonschemaValidator = object  # type: ignore[misc,assignment]

    # Helpful TypeVars for generic classes and functions.
    # These should never be accessed at runtime, only for type checking, so exclude from __all__.
    T = TypeVar("T")
    T_co = TypeVar("T_co", covariant=True)
    T_contra = TypeVar("T_contra", contravariant=True)
else:  # pragma: no cover

    def __getattr__(name: str):
        # Set the BIDSSCHEMATOOLS_DISABLE_RUNTIME_TYPES environment variable to
        # crash out if attributes are ever accessed at runtime.
        # Currently, we do not use this feature.
        global RUNTIME_IMPORT
        try:
            throw = not RUNTIME_IMPORT
        except NameError:
            import os

            RUNTIME_IMPORT = not os.getenv("BIDSSCHEMATOOLS_DISABLE_RUNTIME_TYPES", "")
            if name == "RUNTIME_IMPORT":
                return RUNTIME_IMPORT
            throw = not RUNTIME_IMPORT

        if throw:
            raise AttributeError(
                f"Attribute {name!r} in module {__name__!r} should only be used for type checking."
            )

        if name in __all__:
            if name in _type_map:
                mod, attr = _type_map[name]
            else:
                mod, attr = "typing", name
            globals()[name] = getattr(__import__(mod), attr)
            return globals()[name]

        msg = f"Module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)



from contextlib import nullcontext
from typing import Any, cast

import pytest
from jsonschema.exceptions import SchemaError, ValidationError
from jsonschema.protocols import Validator as JsonschemaValidator
from jsonschema.validators import Draft7Validator, Draft202012Validator

from bidsschematools.utils import jsonschema_validator

DRAFT_7_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {"name": {"type": "string"}},
    "required": ["name"],
}
"""
A minimal valid Draft 7 schema requiring a 'name' property of type 'string'.
"""


DRAFT_202012_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {"title": {"type": "string"}},
    "required": ["title"],
}
"""
A minimal valid Draft 2020-12 schema requiring a 'title' property of type 'string'.
"""

DRAFT_202012_FORMAT_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {"email": {"type": "string", "format": "email"}},
    "required": ["email"],
}
"""
Draft 2020-12 schema that includes a 'format' requirement (e.g., 'email').
Used to test the 'check_format' parameter.
"""


SCHEMA_NO_DOLLAR_SCHEMA = {
    "type": "object",
    "properties": {"foo": {"type": "string"}},
    "required": ["foo"],
}
"""
Schema that lacks the '$schema' property altogether.
Used to test that 'default_cls' is applied.
"""


class TestJsonschemaValidator:
    @pytest.mark.parametrize(
        ("schema", "expected_validator_cls"),
        [
            pytest.param(DRAFT_202012_FORMAT_SCHEMA, Draft202012Validator, id="Draft202012"),
            pytest.param(DRAFT_7_SCHEMA, Draft7Validator, id="Draft7"),
        ],
    )
    @pytest.mark.parametrize("check_format", [True, False])
    def test_set_by_dollar_schema(
        self,
        schema: dict[str, Any],
        expected_validator_cls: type,
        check_format: bool,
    ) -> None:
        """
        Test that the correct validator class is returned for different '$schema' values
        """
        validator = jsonschema_validator(schema, check_format=check_format)

        assert isinstance(validator, expected_validator_cls)

    @pytest.mark.parametrize(
        ("check_format", "instance", "expect_raises"),
        [
            (True, {"email": "test@example.com"}, False),
            (True, {"email": "not-an-email"}, True),
            (False, {"email": "not-an-email"}, False),
        ],
        ids=[
            "check_format=True, valid email",
            "check_format=True, invalid email",
            "check_format=False, invalid email",
        ],
    )
    def test_check_format_email_scenarios(
        self,
        check_format: bool,
        instance: dict,
        expect_raises: bool,
    ) -> None:
        """
        Parametrized test for check_format usage on valid/invalid email addresses under
        Draft202012Validator.
        """
        validator = jsonschema_validator(DRAFT_202012_FORMAT_SCHEMA, check_format=check_format)

        # If expect_raises is True, we use pytest.raises(ValidationError)
        # Otherwise, we enter a no-op context
        ctx = pytest.raises(ValidationError) if expect_raises else nullcontext()

        with ctx:
            validator.validate(instance)  # Should raise or not raise as parametrized

    @pytest.mark.parametrize(
        ("schema", "expected_validator_cls"),
        [
            # Scenario 1: no $schema => we expect the default_cls=Draft7Validator is used
            pytest.param(SCHEMA_NO_DOLLAR_SCHEMA, Draft7Validator, id="no-$schema"),
            # Scenario 2: has $schema => draft 2020-12 overrides the default_cls
            pytest.param(DRAFT_202012_SCHEMA, Draft202012Validator, id="with-$schema"),
        ],
    )
    def test_default_cls(
        self,
        schema: dict[str, Any],
        expected_validator_cls: type,
    ) -> None:
        """
        If the schema has no '$schema' property, and we provide a 'default_cls',
        the returned validator should be an instance of that class.

        If the schema *does* have '$schema', then the default_cls is ignored, and
        the validator class is inferred from the schema's '$schema' field.
        """
        # Provide default_cls=Draft7Validator
        validator = jsonschema_validator(
            schema,
            check_format=False,
            default_cls=cast(type[JsonschemaValidator], Draft7Validator),
        )
        assert isinstance(validator, expected_validator_cls)

    def test_invalid_schema_raises_schema_error(self) -> None:
        """
        Provide an invalid schema, ensuring that 'SchemaError' is raised.
        """
        invalid_schema = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "type": 123,  # 'type' must be string/array, so this is invalid
        }
        with pytest.raises(SchemaError):
            jsonschema_validator(invalid_schema, check_format=False)
